/*---------------------------------------------------------------------------

                       FFFFFFFF  DDDDDD         hh        
                       FF        DD    DD       hh        
                       FFFFFF    DD    DD       hh hhh    
                       FF        DD    DD       hhh   hh  
                       FF        DD    DD  ...  hh    hh  
                       FF        DDDDDD    :::  hh    hh  

  ---------------------------------------------------------------------------

  fd.h  (v3.32)

  Basic header file for inclusion in all programs related to FLOPPY DISC
  code (d###.c, f###.c, repair.c, sh###.c, tesselate.c, update###.c, analy-
  sis programs, etc.).

  ---------------------------------------------------------------------------

     Version:  Y-MP/PC                               Greg Roelofs
     Source:   fd.h on eagle                         25 May 1992

  ---------------------------------------------------------------------------*/



/**************************/
/*  Includes and Defines  */
/**************************/

#include <stdlib.h>      /* rand(), abs(), atoi(), atof() */
#include <stdio.h>       /* printf(), fprintf() */
#include <math.h>        /* pow(), powri(), sin(), cos(), sqrt() */
#include <malloc.h>      /* malloc(), free() */

#define TRUE             1
#define FALSE            0

#define PRETTY_BIG                30.0e5   /* e9 fails sometimes */
#define TRULY_HORRENDOUSLY_BIG    9.9e99

#define E_ON_HULL        0x0001   /* edge is on convex hull */
#define E_UNMODIFIED     0x0002   /* no change:  not candidate for swap */
#define E_FORCE_SWAP     0x0004   /* fold:  point-swap won't work; do edge */
#define E_CHECKED_FOLD   0x0008   /* edge was already checked:  can skip */
#define E_SKIP_VISC      0x0010   /* large fold (or short edge?):  skip calc */

#define UNKNOWN          0        /* all files before v3.10 */
#define POINTS           1        /* file type for d###pts#### */
#define TESS             2        /* file type for d###tes#### */
#define TRI2D            3        /* file type for d###tri#### */
#define TRI3D            4        /* file type for d###h###v###.# */
#define SURF             5        /* file type for d###h###s###.# */
#define FD_DATA          6        /* file type for d###h###t###.# */
#define KARRMOVIE        7        /* file type for d###h###.karr */
#define MOVIE            8        /* file type for d###h###.movie */

#ifndef NOTRI            /* do triangulation as well as tesselation */
#  define TRI
#elif defined(TRI)       /* do tesselation only:  undefine TRI */
#  undef TRI
#endif

#ifdef LOADER            /*================ LOADER equates ===============*/
#  if defined(MSDOS) && defined(TRI)   /* PC, tesselation and triangulation */
#    define NUM_POINTS   90
#  elif defined(MSDOS)   /* PC, tesselation only (max poss. in small model) */
#    define NUM_POINTS   200
#  else /* Y-MP */
#    define NUM_POINTS   4000       /* (approx) interior pts + hull */
#  endif
#  define NUM_DIAM       (NUM_POINTS/30+10)   /* >= sqrt(NUM_INT_POINTS) */
#  define MAX_V          (NUM_POINTS*6/5)
#  define MAX_T          (2*MAX_V)
#  define MAX_E          (3*MAX_V)
#  define MAX_DirV       (MAX_T+15)
#  define HEAP_SIZE      (8*(NUM_POINTS+3))
#else /* !LOADER */      /*=============== EVOLVER equates ===============*/
#  define NUM_POINTS     4800     /* a little extra room... */
#  define MAX_V          (NUM_POINTS+20)
#  define MAX_T          (2*MAX_V)
#  define MAX_E          (3*MAX_V)
#endif                   /*===============================================*/

#define MAX_subT         (MAX_T+NUM_POINTS/10)

#ifdef M_PI
#  define PI             M_PI
#else
#  define PI             3.141592653589793238462643383
#endif
#ifdef M_PI_2
#  define PI_2           M_PI_2
#else
#  define PI_2           (PI/2.)
#endif
#define TWOPI            (2.*PI)
#define PI180            (PI/180.)

#define MAX(a,b)         (((a) > (b)) ? (a) : (b))
#define MIN(a,b)         (((a) < (b)) ? (a) : (b))
#ifdef MSDOS
#  define SQR(x)         ((x)*(x))
#  define powri          pow               /* in case we missed any... */
#else   /* Y-MP */
#  define SQR(x)         powri((x), 2)     /* add fabs()? */
#endif

/*  NOTE:  Must be careful NOT to use any compound expressions in the fol-  */
/*         lowing macros, only simple vectors (V[i].r or s[1], for example, */
/*         which might evaluate to V[i].r[1] and s[1][1], respectively).    */
/*         Spaces don't matter:  (s[1] [1]*s[2] [0]) is OK.                 */

#define XCROSS(v,w)      ((v[1])*(w[2]) - (v[2])*(w[1]))
#define YCROSS(v,w)      ((v[2])*(w[0]) - (v[0])*(w[2]))
#define ZCROSS(v,w)      ((v[0])*(w[1]) - (v[1])*(w[0]))
#define DOT(v,w)         ((v[0])*(w[0]) + (v[1])*(w[1]) + (v[2])*(w[2]))
#define NORM(v)          sqrt(DOT(v, v))
#define DIST(x,y)        sqrt(SQR((x[0])-(y[0])) + SQR((x[1])-(y[1])) + SQR((x[2])-(y[2])))





/***********************************/
/*  "Equivalences" (more defines)  */
/***********************************/

/* variables are uppercase for Cray Fortran COMMON interface */

#define karr          KARR
#define rarr          RARR
#define cm_subT       CMSUBT
#define accel_potent  ACCPOT
#define grav_accel    ACCPOT
#define grav_potent   ACCPOT
#define potent_phantom  accel_phantom

#define d_version     karr[0]    /* reserved for loader version              */
#define iteration     karr[1]    /* step number                              */
#define numE          karr[2]    /* number of edges in triangulation         */
#define numT          karr[3]    /* number of triangles in triangulation     */
#define numV          karr[4]    /* number of vertices in triangulation      */
#define f_version     karr[5]    /* reserved for evolver version             */
#define En            karr[6]    /* shape of potential:  e.g., E3 or E7      */
#define numRings      karr[7]    /* number of (real) rings in points load    */
#define num_phantom   karr[8]    /* number of "phantom wedgies" at periphery */
#define p_version     karr[9]    /* points-load version (really 100*version) */
#define file_type     karr[10]   /* type of data file (ID)                   */
#define karr2size     karr[11]   /* number of elements in karr2[] (if any)   */
#define rarr2size     karr[12]   /* number of elements in rarr2[] (if any)   */
    /* 7 spaces available:  karr[13-19] */
#define pRing_last   &karr[20]   /* (array) index of last vertex in ring     */
/* ring_last[numRings+1]  [20-]     reserved (no more than 50 elements)      */

#define delta_t       rarr[0]    /* time step                                */
#define Time          rarr[1]    /* current time of simulation               */
#define KinE          rarr[2]    /* total kinetic energy                     */
#define GravE         rarr[3]    /* total gravitational energy               */
#define IntE          rarr[4]    /* total internal (heat) energy             */
#define TotE          rarr[5]    /* total energy of disk                     */
#define pLinMom      &rarr[6]    /* linear momentum (vector) of disk         */
/* LinMom[3]          rarr[6-8]     reserved                                 */
#define TotLinMom     rarr[9]    /* linear momentum (magnitude) of disk      */
#define pAngMom      &rarr[10]   /* angular momentum (vector) of disk        */
/* AngMom[3]          rarr[10-12]   reserved                                 */
#define TotAngMom     rarr[13]   /* angular momentum (magnitude) of disk     */
#define pc_length     rarr[14]   /* parsecs per length unit                  */
#define Msol_mass     rarr[15]   /* solar masses per mass unit               */
#define Myr_time      rarr[16]   /* megayears per time unit                  */
#define cm_pc         rarr[17]   /* centimeters per parsec                   */
#define g_Msol        rarr[18]   /* grams per solar mass                     */
#define s_Myr         rarr[19]   /* seconds per megayear                     */
#define G_cgs         rarr[20]   /* gravitational constant in cgs units      */
#define G_pMM         rarr[21]   /* gravitational constant in pc/M_sol/Myr   */
#define G_LMT         rarr[22]   /* gravitational constant in program units  */
#define Rcore         rarr[23]   /* core radius of isothermal potential      */
#define rho0          rarr[24]   /* mass density of halo at origin           */
#define phi0          rarr[25]   /* halo potential at origin                 */
#define c1half        rarr[26]   /* c1 / 2 (oblate Schwarz. potential param) */
#define c2            rarr[27]   /* c2 (potential param, from En)            */
#define Rdisk         rarr[28]   /* disk radius                              */
#define Rholm         rarr[29]   /* radius of optical disk (= unwarped part) */
#define bB            rarr[30]   /* scale factor for expon. surface density  */
#define hole          rarr[31]   /* ratio of hole radius to disk radius (NU) */
#define Mdisk         rarr[32]   /* disk mass                                */
#define mu0           rarr[33]   /* central surface density                  */
#define v_rms         rarr[34]   /* sound speed                              */
#define kT_m          rarr[35]   /* temperature, kT/m (= sound speed^2)      */
#define vis_dyn       rarr[36]   /* dynamic viscosity; assumed constant here */
#define gamma_dyn     rarr[37]   /* gamma defined as:  (bulk - 2/3 shear)    */
#define Inclination   rarr[38]   /* inclination of outer ring wrt inner ring */
#define pRing_last_n &rarr[39]   /* unit normal to outer ring (best fit)     */
/* ring_last_n[3]     rarr[39-41]   reserved                                 */

#define r_cm_phantom  rarr[42]   /* radius of CM of "phantom wedgies"        */
#define mu_phantom_orig rarr[43] /* initial (2D) mass dens of "phantom" ring */
#define mass_phantom  rarr[44]   /* mass of each "phantom wedgie"            */
#define dvdr_phantom  rarr[45]   /* radial velocity gradient of phantom ring */
#define scaleht2_phantom  rarr[46]  /* 2*scale height of "phantom" ring      */
    /* 3 spaces available:  rarr[47-49] */
#define pRing_R      &rarr[50]   /* (array) radii of rings                   */
/* ring_R[numRings+1] rarr[50-]     reserved (no more than 50 elements)      */
/* ring_incl[numRings+1]  [...]     reserved (ditto, contiguous)             */
/* ring_v[numRings+1]     [...]     reserved (ditto)                         */
/* ring_azim[numRings+1]  [...]     reserved (ditto)                         */
    /* only guaranteed 6 spaces at end:  rarr[250-255] */
/*                    rarr[255]     used (temporarily) between p### and d### */





/*****************************/
/*  Typedefs and Prototypes  */
/*****************************/

typedef double   vector[3];
typedef double   four_vector[4];

/*
 * Loader routines
 */

void             construct_points       ( void );
void             tesselate              ( void );
void             triangulate            ( void );
void             physical_variables     ( void );
double           el2                    ( double, double, double, double );
void             write_points           ( void );
void             write_tess             ( void );
void             initialize_plot        ( void );
void             plot_points            ( void );
void             plot_tiles             ( int );
void             plot_tri               ( void );
void             draw_line              ( double, double, double, double, 
                                          double * );


/*
 * Evolver routines
 */

void             read_data              ( void );
void             evolve                 ( void );
void             update_points          ( void );
void             repair_grid            ( void );
void             swap_points            ( void );
void             write_movie_data       ( int );
void             nrerror                ( char * );
void             grr_jacobi             ( double **, int, double *,
                                          double **, int * );
fortran int      ISMIN                  ( int *, double *, int * );


/*
 * Shared routines
 */

int              swap_edges             ( int );      /* sh###.c */
void             print_phys_vars        ( void );     /* sh###.c */
void             write_edges            ( void );     /* sh###.c */
void             write_tri              ( void );     /* sh###.c */
void             write_data             ( void );     /* sh###.c */
fortran void     GRAV                   ( int * );               /* GRR: 290 */
fortran void     VAXWR                  ( int *, double *, double *, int * );
fortran void     VAXWR3                 ( int *, double *, double *, double *,
                                          double *, int *, int *, int * );
fortran void     FORWR3                 ( int *, double *, double *, double *,
                                          double *, int *, int *, int * );
fortran double   SECOND                 ( void );





/************/
/*  Macros  */
/************/

#define MALLOC_OPEN_WRITEPTS(n) \
   {coord=(double *)malloc((unsigned)(n)*sizeof(double)); if(!coord) {printf(\
   "error:  couldn't allocate memory to write [ %s ]\n",filename); return;}\
   if((tessfile=fopen(filename,"wb"))==NULL) {printf(\
   "error:  couldn't open file [ %s ]\n",filename); free(coord); return;}\
   total_bytes=fwrite(karr,sizeof(int),256,tessfile)*sizeof(int);\
   total_bytes+=fwrite(rarr,sizeof(double),256,tessfile)*sizeof(double);\
   for(p=coord,i=num_fake;i<numPt;++i) *p++=Pt[i].x;\
   for(i=num_fake;i<numPt;++i) *p++=Pt[i].y; total_bytes+=fwrite(coord,\
   sizeof(double),2*numV,tessfile)*sizeof(double);}

/*
 *  Malloc buffer into which to copy coordinates, open output file, write
 *  karr[] and rarr[], copy the coordinates of the points to the buffer,
 *  and write that.  This macro is used by write_points() in p###.c and
 *  by write_tess() in d###.c.
 *
 *  macro MALLOC_OPEN_WRITEPTS(n)
 *      int  n;
 *  {
 *      coord = (double *) malloc((unsigned) (n)*sizeof(double));
 *      if (!coord) {
 *          printf("error:  couldn't allocate memory to write [ %s ]\n",
 *            filename);
 *          return;
 *      }
 *      if ((tessfile = fopen(filename,"wb")) == NULL) {
 *          printf("error:  couldn't open file [ %s ]\n", filename);
 *          free(coord);
 *          return;
 *      }
 *      total_bytes = fwrite(karr, sizeof(int), 256, tessfile) * sizeof(int);
 *      total_bytes += fwrite(rarr, sizeof(double), 256, tessfile) *
 *        sizeof(double);
 *  
 *      for (p = coord, i = num_fake;  i < numPt;  ++i)
 *          *p++ = Pt[i].x;
 *      for (i = num_fake;  i < numPt;  ++i)
 *          *p++ = Pt[i].y;
 *  
 *                   /-*   this should NOT be (n)!!! ---v   *-/
 *      total_bytes += fwrite(coord, sizeof(double), 2*numV, tessfile) *
 *                                   sizeof(double);
 *  }
 *
 */





/************************/
/*  External Variables  */
/************************/

/*
 * Loader variables
 */

extern int       num_DelDirV, max_DelDirV, num_ContigPt, max_ContigPt;
extern int       numDirV, numPt, num_compactions, Eflag;
extern int       num_fake, heap_avail;
extern double    tessel_time;

extern struct POINT {
    double x;
    double y;
    int num_contig;
    struct POINT **pHeap;
    struct DVERTEX *pDirV;
}                Pt[], *Heap[], *pNewPt, *pContigPt[], **pBottomHeap;

extern struct DVERTEX {
    double x;
    double y;
    double rsqr;
    struct POINT *pFormPt[3];
    struct DVERTEX *pNbrDirV[3];
}                DirV[], *pDelDirV[], *pNewDirV[], *pPenumDirV[];

/*
 * Evolver variables
 */

#ifndef LOADER
   extern double     grav_time;
#endif
extern double        update_time, jacobi_time, repair_time, swappts_time;
extern vector        *cm_phantom;
extern four_vector   *accel_phantom;
extern FILE          *infile, *dbgfile, *moviefile, *moviejunk;

/*
 * Shared variables
 */

extern char          filename[];
extern int           karr[], errorlevel, num_swaps, *const ring_last;
extern double        rarr[], *const LinMom, *const AngMom;
extern double        *const ring_last_n, *const ring_R;
extern double        *ring_incl, *ring_v, *ring_azim;    /* all three vary */
extern double        swapE_time, write_time;
extern vector        s[], cm_subT[][3];
extern four_vector   accel_potent[][3];
extern FILE          *outfile;

extern struct edge {
    struct triangle *pT[2];                      /* bottom of triangulate()  */
    struct vertex   *pV[2];                      /* bottom of triangulate()  */
    double *n[2];   /* ptrs to normals in T[] */ /* end physical_variables() */
    double len;                                  /* bottom of triangulate()  */
    int status;                                  /* triangulate()            */
}                E[];

extern struct triangle {
    struct vertex *pV[3];                        /* top of triangulate()     */
    struct edge   *pE[3];                        /* bottom of triangulate()  */
    vector n[3];   /* unit normals to 3 edges */ /* physical_variables()     */
    double mass;                                 /* physical_variables()     */
    double scaleht2;   /* scale ht. * 2 */       /* physical_variables()     */
    double pressure;                             /* not initialized          */
 /* double temperature; */
 /* double internal_E;  */
 /* double ion_frac;    */
}                T[];

extern struct vertex {
    vector r;                                    /* top of triangulate()     */
    vector v;                                    /* physical_variables()     */
    vector F;                                    /* physical_variables()     */
    double mass;                                 /* physical_variables()     */
    int attrib;                                  /* top of triangulate()     */
}                V[];
